# SPDX-FileCopyrightText: 2021 GNOME Foundation
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import argparse
import json
import os
import sys

from . import gir, log, utils


HELP_MSG = "Search terms in the symbol indices"


def _gen_alias_result(symbol, namespace):
    alias = namespace.find_alias(symbol["name"])
    if alias.doc is not None:
        summary = utils.preprocess_docs(alias.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "alias",
        "name": alias.name,
        "ctype": alias.base_ctype,
        "summary": summary,
        "link": f"alias.{alias.name}.html",
    }


def _gen_bitfield_result(symbol, namespace):
    enum = namespace.find_bitfield(symbol["name"])
    if enum.doc is not None:
        summary = utils.preprocess_docs(enum.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "flags",
        "name": enum.name,
        "ctype": enum.base_ctype,
        "summary": summary.split("\n"),
        "link": f"flags.{enum.name}.html",
    }


def _gen_callback_result(symbol, namespace):
    pass


def _gen_class_result(symbol, namespace):
    cls = namespace.find_class(symbol["name"])
    if cls.doc is not None:
        summary = utils.preprocess_docs(cls.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "class",
        "name": cls.name,
        "ctype": cls.base_ctype,
        "summary": summary.split("\n"),
        "link": f"class.{cls.name}.html",
    }


def _gen_class_method_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])
    cls_name = t.struct_for
    if cls_name is None:
        return None

    methods = getattr(t, "methods", [])
    for m in methods:
        if m.name == symbol["name"]:
            if m.doc is not None:
                summary = utils.preprocess_docs(m.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"

            return {
                "type": "class method",
                "name": f"{t.name}.{m.name}",
                "ctype": m.identifier,
                "summary": summary.split("\n"),
                "link": f"class_method.{cls_name}.{m.name}.html",
            }

    return None


def _gen_constant_result(symbol, namespace):
    const = namespace.find_real_type(symbol["name"])

    if const.doc is not None:
        summary = utils.preprocess_docs(const.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "constant",
        "name": const.name,
        "ctype": const.ctype,
        "summary": summary.split("\n"),
        "link": f"const.{const.name}.html",
    }


def _gen_ctor_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    methods = getattr(t, "constructors", [])
    for m in methods:
        if m.name == symbol["name"]:
            if m.doc is not None:
                summary = utils.preprocess_docs(m.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"

            return {
                "type": "constructor",
                "name": f"{t.name}.{m.name}",
                "ctype": m.identifier,
                "summary": summary.split("\n"),
                "link": f"ctor.{t.name}.{m.name}.html",
            }

    return None


def _gen_domain_result(symbol, namespace):
    enum = namespace.find_error_domain(symbol["name"])
    if enum.doc is not None:
        summary = utils.preprocess_docs(enum.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "error",
        "name": enum.name,
        "ctype": enum.base_ctype,
        "summary": summary.split("\n"),
        "link": f"error.{enum.name}.html",
    }


def _gen_enum_result(symbol, namespace):
    enum = namespace.find_enumeration(symbol["name"])
    if enum.doc is not None:
        summary = utils.preprocess_docs(enum.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "enum",
        "name": enum.name,
        "ctype": enum.base_ctype,
        "summary": summary.split("\n"),
        "link": f"enum.{enum.name}.html",
    }


def _gen_func_result(symbol, namespace):
    functions = namespace.get_functions()
    for func in functions:
        if func.name == symbol["name"]:
            if func.doc is not None:
                summary = utils.preprocess_docs(func.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"
            return {
                "type": "function",
                "name": func.name,
                "ctype": func.identifier,
                "summary": summary.split("\n"),
                "link": f"func.{func.name}.html",
            }


def _gen_func_macro_result(symbol, namespace):
    macros = namespace.get_effective_function_macros()
    for func in macros:
        if func.name == symbol["name"]:
            if func.doc is not None:
                summary = utils.preprocess_docs(func.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"
            return {
                "type": "function macro",
                "name": func.name,
                "ctype": func.identifier,
                "summary": summary.split("\n"),
                "link": f"func.{func.name}.html",
            }


def _gen_interface_result(symbol, namespace):
    iface = namespace.find_interface(symbol["name"])
    if iface.doc is not None:
        summary = utils.preprocess_docs(iface.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "interface",
        "name": iface.name,
        "ctype": iface.base_ctype,
        "summary": summary.split("\n"),
        "link": f"iface.{iface.name}.html",
    }


def _gen_method_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    methods = getattr(t, "methods", [])
    for m in methods:
        if m.name == symbol["name"]:
            if m.doc is not None:
                summary = utils.preprocess_docs(m.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"

            return {
                "type": "method",
                "name": f"{t.name}.{m.name}",
                "ctype": m.identifier,
                "summary": summary.split("\n"),
                "link": f"method.{t.name}.{m.name}.html",
            }

    return None


def _gen_property_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    properties = getattr(t, "properties", {})
    prop = properties.get(symbol["name"], None)
    if prop is None:
        return None

    if prop.doc is not None:
        summary = utils.preprocess_docs(prop.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "property",
        "name": f"{t.name}:{prop.name}",
        "ctype": t.base_ctype,
        "summary": summary.split("\n"),
        "link": f"property.{t.name}.{prop.name}.html",
    }


def _gen_record_result(symbol, namespace):
    record = namespace.find_record(symbol["name"])
    if record.doc is not None:
        summary = utils.preprocess_docs(record.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "struct",
        "name": record.name,
        "ctype": record.base_ctype,
        "summary": summary.split("\n"),
        "link": f"struct.{record.name}.html",
    }


def _gen_signal_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    signals = getattr(t, "signals", {})
    signal = signals.get(symbol["name"], None)
    if signal is None:
        return None

    if signal.doc is not None:
        summary = utils.preprocess_docs(signal.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "signal",
        "name": f"{t.name}::{signal.name}",
        "ctype": t.base_ctype,
        "summary": summary.split("\n"),
        "link": f"signal.{t.name}.{signal.name}.html",
    }


def _gen_type_func_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    functions = getattr(t, "functions", [])
    for f in functions:
        if f.name == symbol["name"]:
            if f.doc is not None:
                summary = utils.preprocess_docs(f.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"

            return {
                "type": "func",
                "name": f"{t.name}.{f.name}",
                "ctype": f.identifier,
                "summary": summary.split("\n"),
                "link": f"type_func.{t.name}.{f.name}.html",
            }

    return None


def _gen_union_result(symbol, namespace):
    union = namespace.find_union(symbol["name"])
    if union.doc is not None:
        summary = utils.preprocess_docs(union.doc.content, namespace, summary=True, plain=True)
    else:
        summary = "Missing documentation"

    return {
        "type": "union",
        "name": union.name,
        "ctype": union.base_ctype,
        "summary": summary.split("\n"),
        "link": f"union.{union.name}.html",
    }


def _gen_vfunc_result(symbol, namespace):
    t = namespace.find_real_type(symbol["type_name"])

    methods = getattr(t, "virtual_methods", [])
    for m in methods:
        if m.name == symbol["name"]:
            if m.doc is not None:
                summary = utils.preprocess_docs(m.doc.content, namespace, summary=True, plain=True)
            else:
                summary = "Missing documentation"

            return {
                "type": "vfunc",
                "name": f"{t.name}.{m.name}",
                "ctype": t.base_ctype,
                "summary": summary.split("\n"),
                "link": f"vfunc.{t.name}.{m.name}.html",
            }

    return None


def search_symbols(real_term, symbols):
    term = real_term.lower()

    matches = []
    for symbol in symbols:
        if any([
            term in symbol.get("name", "").lower(),
            term in symbol.get("ctype", "").lower(),
            term in symbol.get("summary", "").lower(),
        ]):
            matches.append(symbol)
    return matches


def query(repository, terms, index_file):
    if index_file is None:
        index_file = os.path.join(os.getcwd(), "index.json")

    with open(index_file, "r", encoding="utf-8") as f:
        index = json.load(f)

    namespace = repository.namespace

    index_meta = index["meta"]
    index_symbols = index["symbols"]

    if index_meta["ns"] != namespace.name or index_meta["version"] != namespace.version:
        log.error("Index file does not match the GIR namespace")

    result_types = {
        "alias": _gen_alias_result,
        "bitfield": _gen_bitfield_result,
        "callback": _gen_callback_result,
        "class": _gen_class_result,
        "class_method": _gen_class_method_result,
        "constant": _gen_constant_result,
        "ctor": _gen_ctor_result,
        "domain": _gen_domain_result,
        "enum": _gen_enum_result,
        "function": _gen_func_result,
        "function_macro": _gen_func_macro_result,
        "interface": _gen_interface_result,
        "method": _gen_method_result,
        "property": _gen_property_result,
        "record": _gen_record_result,
        "signal": _gen_signal_result,
        "type_func": _gen_type_func_result,
        "union": _gen_union_result,
        "vfunc": _gen_vfunc_result,
    }

    results = []

    for term in terms:
        symbols = search_symbols(term, index_symbols)
        for symbol in symbols:
            gen_result = result_types.get(symbol["type"])
            if gen_result is None:
                log.warning(f"Unhandled symbol type {symbol['type']} for '{term}'")
                continue

            res = gen_result(symbol, namespace)
            if not res:
                log.debug(f"No details for result {symbol['name']}")
            else:
                log.debug(f"Adding results for {symbol['name']}")
                results.append(res)

    prefix = "result:"
    indent = ' ' * len(prefix)

    n_results = len(results)
    terms_str = ", ".join(terms)
    print(f"Found {n_results} results matching '{terms_str}'")

    for res in results:
        lines = [f"{prefix} [{res['type']}] {res['name']} - {res['ctype']}"]
        for chunk in res["summary"]:
            lines.append(f"{indent} {chunk}")
        lines.append(f"{indent} link: {res['link']}")
        print("\n".join(lines))


def add_args(parser):
    parser.add_argument("--add-include-path", action="append", dest="include_paths", default=[],
                        help="include paths for other GIR files")
    parser.add_argument("--index", help="the index file")
    parser.add_argument("--term", action="append", dest="terms", default=[],
                        help="a search term")
    parser.add_argument("infile", metavar="GIRFILE", type=argparse.FileType('r', encoding='UTF-8'),
                        default=sys.stdin, help="the GIR file to parse")


def run(options):
    paths = []
    paths.extend(options.include_paths)
    paths.append(utils.default_search_paths())
    log.debug(f"Search paths: {paths}")

    log.info("Parsing GIR file")
    parser = gir.GirParser(search_paths=paths)
    parser.parse(options.infile)

    log.checkpoint()
    query(parser.get_repository(), options.terms, options.index)

    return 0



# SPDX-FileCopyrightText: 2021 GNOME Foundation
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import markdown
import os
import re
import subprocess
import sys

from markupsafe import Markup
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter
from typogrify.filters import typogrify

from . import gir, log, mdext


# The beginning of a gtk-doc code block:
#
# |[ (optional language identifier)
#
CODEBLOCK_START_RE = re.compile(
    r'''
    ^
    \s*
    \|\[
    \s*
    (?P<language>\<\!-- \s* language="\w+" \s* --\>)?
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

# The optional language identifier for a gtk-doc code block:
#
# <!-- language="..." -->
#
LANGUAGE_RE = re.compile(
    r'''
    ^
    \s*
    <!--
    \s*
    language="(?P<language>\w+)"
    \s*
    -->
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

# The ending of a gtk-doc code block:
#
# ]|
#
CODEBLOCK_END_RE = re.compile(
    r'''
    ^
    \s*
    \]\|
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

LINK_RE = re.compile(
    r'''
    (?P<text>\[ [\w\s,\-_:]+ \])?
    \[
    (`)?
    (?P<fragment>[\w]+)
    @
    (?P<endpoint>[\w\-_:\.]+)
    (`)?
    \]
    ''',
    re.VERBOSE)

TYPE_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    ''',
    re.VERBOSE)

PROPERTY_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    :{1}                        # delimiter
    (?P<property>[\w-]*\w)      # property name
    ''',
    re.VERBOSE)

SIGNAL_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    :{2}                        # delimiter
    (?P<signal>[\w-]*\w)        # signal name
    ''',
    re.VERBOSE)

METHOD_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    \.                          # delimiter
    (?P<method>[\w_]*\w)        # method name
    ''',
    re.VERBOSE)

CAMEL_CASE_START_RE = re.compile(r"([A-Z]+)([A-Z][a-z])")

CAMEL_CASE_CHUNK_RE = re.compile(r"([a-z\d])([A-Z])")

LANGUAGE_MAP = {
    'c': 'c',
    'css': 'css',
    'plain': 'plain',
    'xml': 'xml',
    'javascript': 'javascript',
}

MD_EXTENSIONS = [
    # Standard extensions
    'codehilite',
    'def_list',
    'fenced_code',
    'meta',
    'tables',
    'toc',

    # Local extensions
    mdext.GtkDocExtension(),
    mdext.AdmonitionExtension(),
]

MD_EXTENSIONS_CONF = {
    'codehilite': {'guess_lang': False},
    'toc': {'permalink_class': 'md-anchor', 'permalink': ''},
}

EN_STOPWORDS = set("""
a  and  are  as  at
be  but  by
for
if  in  into  is  it
near  no  not
of  on  or
such
that  the  their  then  there  these  they  this  to
was  will  with
""".split())


def process_language(lang):
    if lang is None:
        return "plain"

    res = LANGUAGE_RE.match(lang)
    if res:
        language = res.group("language") or "plain"
    else:
        language = "plain"

    return LANGUAGE_MAP[language.lower()]


class LinkParseError:
    def __init__(self, line=None, start=0, end=0, fragment=None, rest=None, message="Unable to parse link"):
        self.line = line
        self.start = start
        self.end = end
        self.fragment = fragment
        self.rest = rest
        self.message = message

    def __str__(self):
        if self.line is not None:
            msg = [self.message]
            msg.append(self.line)
            err_line = ['^'.rjust(self.start + 1, ' ')]
            err_line += [''.join(['~' for x in range(self.end - self.start - 1)])]
            msg.append("".join(err_line))
            return "\n".join(msg)
        else:
            return f"{self.message}: [{self.fragment}@{self.rest}]"


class LinkGenerator:
    def __init__(self, **kwargs):
        self._line = kwargs.get('line')
        self._start = kwargs.get('start', 0)
        self._end = kwargs.get('end', 0)
        self._namespace = kwargs.get('namespace')
        self._fragment = kwargs.get('fragment', '')
        self._endpoint = kwargs.get('endpoint', '')
        self._no_link = kwargs.get('no_link', False)
        self._alt_text = kwargs.get('text')

        assert self._namespace is not None

        self._repository = self._namespace.repository
        self._valid_namespaces = [n for n in self._repository.includes]
        self._external = False

        fragment_parsers = {
            "alias": self._parse_type,
            "callback": self._parse_type,
            "class": self._parse_type,
            "const": self._parse_type,
            "ctor": self._parse_method,
            "enum": self._parse_type,
            "error": self._parse_type,
            "flags": self._parse_type,
            "func": self._parse_func,
            "id": self._parse_id,
            "iface": self._parse_type,
            "method": self._parse_method,
            "property": self._parse_property,
            "signal": self._parse_signal,
            "struct": self._parse_type,
            "type": self._parse_type,
            "vfunc": self._parse_method,
        }

        parser_method = fragment_parsers.get(self._fragment)
        if parser_method is not None:
            res = parser_method(self._fragment)
            if res is not None:
                self._fragment = None
                log.warning(str(res))
        else:
            self._fragment = None
            log.warning(str(LinkParseError(self._line, self._start, self._end,
                                           self._fragment, self._endpoint,
                                           "Unable to parse link")))

    def _parse_id(self, fragment):
        symbol = self._repository.find_symbol(self._endpoint)
        if symbol is None:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find symbol {self._endpoint}")
        (ns, t) = symbol
        if isinstance(t, gir.Class) or \
           isinstance(t, gir.Interface) or \
           isinstance(t, gir.Record):
            self._external = ns is not self._namespace
            self._ns = ns.name
            self._fragment = 'method'
            self._symbol_name = f"{self._endpoint}()"
            self._name = t.name
            self._method_name = self._endpoint.replace(ns.symbol_prefix[0] + '_', '')
            self._method_name = self._method_name.replace(t.symbol_prefix + '_', '')
            return None
        elif isinstance(t, gir.Function):
            self._external = ns is not self._namespace
            self._ns = ns.name
            self._fragment = 'func'
            self._symbol_name = f"{self._endpoint}()"
            self._name = None
            self._func_name = self._endpoint.replace(ns.symbol_prefix[0] + '_', '')
            return None
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unsupported symbol {self._endpoint}")

    def _parse_type(self, fragment):
        res = TYPE_RE.match(self._endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  "Invalid type link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                self._fragment = None
                return None
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            # We determine the fragment here, in case `type` was used,
            # or for validating the fragment passed, to avoid creating
            # invalid links
            if isinstance(t, gir.Alias):
                type_fragment = 'alias'
            elif isinstance(t, gir.Callback):
                type_fragment = 'callback'
            elif isinstance(t, gir.Class):
                type_fragment = 'class'
            elif isinstance(t, gir.Constant):
                type_fragment = 'const'
            elif isinstance(t, gir.Enumeration):
                if isinstance(t, gir.BitField):
                    type_fragment = 'flags'
                elif isinstance(t, gir.ErrorDomain):
                    type_fragment = 'error'
                else:
                    type_fragment = 'enum'
            elif isinstance(t, gir.Interface):
                type_fragment = 'iface'
            elif isinstance(t, gir.Record) or isinstance(t, gir.Union):
                type_fragment = 'struct'
            else:
                return LinkParseError(self._line, self._start, self._end,
                                      self._fragment, self._endpoint,
                                      f"Invalid type {t} for '{ns}.{name}'")
            if fragment != 'type' and fragment != type_fragment:
                return LinkParseError(self._line, self._start, self._end,
                                      self._fragment, self._endpoint,
                                      f"Invalid fragment for '{ns}.{name}': it should be {type_fragment}")
            self._fragment = type_fragment
            self._name = name
            self._type = t.base_ctype
            return None
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find type '{ns}.{name}'")

    def _parse_property(self, fragment):
        res = PROPERTY_RE.match(self._endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            pname = res.group('property')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
            # Canonicalize the property name
            pname = pname.replace('_', '-')
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  "Invalid property link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = ns
            else:
                self._fragment = None
                return None
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._name = name
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find type '{ns}.{name}'")
        if (isinstance(t, gir.Class) or isinstance(t, gir.Interface)) and pname in t.properties:
            self._property_name = pname
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Invalid property '{pname}' for type '{ns}.{name}'")

    def _parse_signal(self, fragment):
        res = SIGNAL_RE.match(self._endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            sname = res.group('signal')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
            # Canonicalize the signal name
            sname = sname.replace('_', '-')
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  "Invalid signal link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                self._fragment = None
                return None
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._name = name
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find type '{ns}.{name}'")
        if (isinstance(t, gir.Class) or isinstance(t, gir.Interface)) and sname in t.signals:
            self._signal_name = sname
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Invalid signal name '{sname}' for type '{ns}.{name}'")

    def _parse_method(self, fragment):
        res = METHOD_RE.match(self._endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            method = res.group('method')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  "Invalid method link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._ns = namespace.name
                self._external = True
            else:
                self._fragment = None
                return None
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._method_name = method
            # method@Foo.BarClass.add_name -> class_method.Bar.add_name.html
            if isinstance(t, gir.Record) and t.struct_for is not None:
                self._name = t.struct_for
                self._fragment = "class_method"
            elif fragment == "vfunc" and t.type_struct is not None:
                self._name = name
                self._type = t.type_struct
            else:
                self._name = name
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find type '{ns}.{name}'")
        if fragment == "ctor":
            methods = getattr(t, "constructors", [])
        elif fragment in ["method", "class_method"]:
            methods = getattr(t, "methods", [])
        elif fragment == "vfunc":
            methods = getattr(t, "virtual_methods", [])
        else:
            methods = []
        for m in methods:
            if m.name == method:
                if fragment == "vfunc":
                    self._vfunc_name = m.name
                else:
                    self._symbol_name = f"{m.identifier}()"
                return None
        return LinkParseError(self._line, self._start, self._end,
                              self._fragment, self._endpoint,
                              f"Unable to find method '{ns}.{name}.{method}'")

    def _parse_func(self, fragment):
        tokens = self._endpoint.split('.')
        # Case 1: [func@init] => gtk_init()
        if len(tokens) == 1:
            ns = self._namespace.name
            name = None
            func_name = tokens[0]
        # Case 2: [func@Gtk.Foo.bar] => gtk_foo_bar()
        elif len(tokens) == 3:
            ns = tokens[0]
            name = tokens[1]
            func_name = tokens[2]
        # Case 3: either [func@Gtk.init] or [func@Foo.bar]
        elif len(tokens) == 2:
            if tokens[0] == self._namespace.name:
                ns = tokens[0]
                name = None
                func_name = tokens[1]
            elif tokens[0] in self._valid_namespaces:
                ns = tokens[0]
                name = None
                func_name = tokens[1]
            else:
                ns = self._namespace.name
                name = tokens[0]
                func_name = tokens[1]
        else:
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  "Invalid function link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                self._fragment = None
                log.warning(f"Namespace {ns} not found for link {self._endpoint}")
                return None
        if name is None:
            t = namespace.find_function(func_name)
            if t is not None:
                self._name = None
                self._func_name = func_name
                self._symbol_name = f"{t.identifier}()"
                return None
            else:
                return LinkParseError(self._line, self._start, self._end,
                                      self._fragment, self._endpoint,
                                      f"Unable to find function '{ns}.{func_name}'")
        else:
            t = namespace.find_real_type(name)
            if t is None:
                return LinkParseError(self._line, self._start, self._end,
                                      self._fragment, self._endpoint,
                                      f"Unable to find type '{ns}.{name}'")
            for func in t.functions:
                if func.name == func_name:
                    self._name = name
                    self._func_name = func.name
                    self._symbol_name = f"{func.identifier}()"
                    return None
            return LinkParseError(self._line, self._start, self._end,
                                  self._fragment, self._endpoint,
                                  f"Unable to find function '{ns}.{name}.{func_name}'")

    @property
    def text(self):
        if self._alt_text is not None:
            return self._alt_text[1:len(self._alt_text) - 1]
        elif self._fragment in ['alias', 'callback', 'class', 'const', 'enum', 'error', 'flags', 'iface', 'struct']:
            return f"<code>{self._type}</code>"
        elif self._fragment == 'property':
            return f"<code>{self._type}:{self._property_name}</code>"
        elif self._fragment == 'signal':
            return f"<code>{self._type}::{self._signal_name}</code>"
        elif self._fragment in ['ctor', 'func', 'method', 'class_method']:
            return f"<code>{self._symbol_name}</code>"
        elif self._fragment == 'vfunc':
            return f"<code>{self._ns}.{self._type}.{self._vfunc_name}</code>"
        else:
            return f"{self._endpoint}"

    @property
    def href(self):
        if self._fragment in ['alias', 'callback', 'class', 'const', 'enum', 'error', 'flags', 'iface', 'struct']:
            return f"{self._fragment}.{self._name}.html"
        elif self._fragment == 'property':
            return f"property.{self._name}.{self._property_name}.html"
        elif self._fragment == 'signal':
            return f"signal.{self._name}.{self._signal_name}.html"
        elif self._fragment in ['ctor', 'method', 'class_method', 'vfunc']:
            return f"{self._fragment}.{self._name}.{self._method_name}.html"
        elif self._fragment == 'func':
            if self._name is not None:
                return f"type_func.{self._name}.{self._func_name}.html"
            else:
                return f"func.{self._func_name}.html"
        else:
            return None

    def __str__(self):
        text = self.text
        if self._no_link:
            return text
        link = self.href
        if link is None:
            return text
        if self._external:
            data_namespace = f"data-namespace=\"{self._ns}\""
            data_link = f"data-link=\"{link}\""
            href = "href=\"javascript:void(0)\""
            css = "class=\"external\""
            return f"<a {href} {data_namespace} {data_link} {css}>{text}</a>"
        else:
            return f"<a href=\"{link}\">{text}</a>"


def preprocess_docs(text, namespace, summary=False, md=None, extensions=[], plain=False, max_length=20):
    if plain:
        text = text.replace('\n', ' ')
        text = re.sub(r'<[^<]+?>', '', text)
        if max_length > 0:
            words = text.split(' ')
            if len(words) > max_length:
                words = words[:max_length - 1]
                words.append('...')
                text = ' '.join(words)
        return text

    processed_text = []

    code_block_text = []
    code_block_language = None
    inside_code_block = False

    for line in text.split("\n"):
        # If we're in "summary" mode, we bail out at the first empty line
        # after a paragraph
        if summary and line == '' and len(processed_text) > 0:
            break

        res = CODEBLOCK_START_RE.match(line)
        if res:
            code_block_language = process_language(res.group("language"))
            inside_code_block = True
            continue

        res = CODEBLOCK_END_RE.match(line)
        if res and inside_code_block:
            if code_block_language == "plain":
                processed_text += ["```"]
                processed_text.extend(code_block_text)
                processed_text += ["```"]
            else:
                lexer = get_lexer_by_name(code_block_language)
                formatter = HtmlFormatter()
                code_block = highlight("\n".join(code_block_text), lexer, formatter)
                processed_text += [""]
                processed_text.extend(code_block.split("\n"))
                processed_text += [""]

            code_block_language = None
            code_block_text = []
            inside_code_block = False
            continue

        if inside_code_block:
            code_block_text.append(line)
        else:
            new_line = []
            idx = 0
            for m in LINK_RE.finditer(line, idx):
                fragment = m.group('fragment')
                endpoint = m.group('endpoint')
                text = m.group('text')
                start = m.start()
                end = m.end()
                link = LinkGenerator(line=line, start=start, end=end,
                                     namespace=namespace,
                                     fragment=fragment, endpoint=endpoint,
                                     no_link=summary, text=text)
                left_pad = line[idx:start]
                replacement = re.sub(LINK_RE, str(link), line[start:end])
                new_line.append(left_pad)
                new_line.append(replacement)
                idx = end
            new_line.append(line[idx:])

            if len(new_line) == 0:
                processed_text.append(line)
            else:
                processed_text.append("".join(new_line))

    if len(processed_text) == 0:
        return ''

    # Capitalize the first character of the first line, but only if it does not
    # start with a link or a gtk-doc marker, to avoid messing up the rest of
    # the string
    first_line = processed_text[0]
    if first_line and first_line[0].isalpha():
        processed_text[0] = ''.join([first_line[0:1].upper(), first_line[1:]])

    # Append a period, if one isn't there already
    last_line = processed_text[-1]
    if last_line and last_line[-1].isalpha():
        processed_text[-1] = ''.join([last_line, '.'])

    if md is None:
        md_ext = extensions.copy()
        md_ext.extend(MD_EXTENSIONS)
        text = markdown.markdown("\n".join(processed_text),
                                 extensions=md_ext,
                                 extension_configs=MD_EXTENSIONS_CONF)
    else:
        text = md.reset().convert("\n".join(processed_text))

    return Markup(typogrify(text, ignore_tags=['h1', 'h2', 'h3', 'h4']))


def code_highlight(text, language='c'):
    lexer = get_lexer_by_name(language)
    formatter = HtmlFormatter()
    return Markup(highlight(text, lexer, formatter))


def render_dot(dot, output_format="svg"):
    if output_format not in ["svg", "png"]:
        log.error("Invalid output format for render_dot(): {output_format}")

    args = ["dot", f"-T{output_format}"]

    try:
        proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.stdin.write(dot.encode("utf-8"))
        output, err = proc.communicate()
        if err:
            log.warning(f"Unable to process dot data: {err}")
            return None
        if output_format == "svg":
            return output.decode("utf-8")
    except Exception as e:
        log.warning(f"Unable to process dot data: {e}")
        return None


found_programs = {}


def find_program(bin_name, path=None):
    """Finds a program @bin_name inside the given @path, and returns
    its full path if found, or None if the program could not be found.

    The @bin_name will automatically get an extension depending on the
    platform.
    """
    global found_programs

    if path is None and bin_name in found_programs:
        return found_programs[bin_name]

    if path is None:
        search_paths = os.environ['PATH'].split(os.pathsep)
    else:
        search_paths = path.split(os.pathsep)

    bin_extensions = ['']

    if sys.platform == 'win32':
        pathext = os.environ['PATHEXT'].lower().split(os.pathsep)
        (basename, extension) = os.path.splitext(bin_name)
        if extension.lower() not in pathext:
            bin_extensions = pathext
        search_paths.insert(0, '')

    for ext in bin_extensions:
        executable = bin_name + ext

        for p in search_paths:
            full_path = os.path.join(p, executable)
            if os.path.isfile(full_path):
                # Memoize the result with the default PATH, so we can
                # call this multiple times at no additional cost
                if path is None:
                    found_programs[bin_name] = full_path
                return full_path

    return None


def default_search_paths():
    if not sys.platform == 'win32':
        xdg_data_dirs = os.environ.get("XDG_DATA_DIRS", "/usr/share:/usr/local/share").split(":")
        xdg_data_home = os.environ.get("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
    else:
        xdg_data_dirs = None
        xdg_data_home = None

    paths = []
    paths.append(os.getcwd())
    gi_gir_path = os.environ.get('GI_GIR_PATH')
    if gi_gir_path is not None:
        paths.extend(gi_gir_path.split(os.pathsep))
    # Add sys.base_prefix when using MSYS2
    if sys.platform == 'win32' and 'GCC' in sys.version:
        paths.append(os.path.join(sys.base_prefix, 'share', 'gir-1.0'))
    if xdg_data_home is not None:
        paths.append(os.path.join(xdg_data_home, "gir-1.0"))
    if xdg_data_dirs is not None:
        paths.extend([os.path.join(x, "gir-1.0") for x in xdg_data_dirs])
    if sys.platform != 'win32' and '/usr/share/gir-1.0' not in paths:
        paths.append('/usr/share/gir-1.0')

    return paths


def find_extra_content_file(content_dirs, file):
    for p in content_dirs:
        full_path = os.path.join(p, file)
        if os.path.isfile(full_path):
            return full_path

    raise FileNotFoundError(f"Content file {file} not found in any content directory")



# SPDX-FileCopyrightText: 2022  Emmanuele Bassi
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import os
import posixpath
import urllib

from http import server

try:
    from magic import from_file as magic_from_file
except ImportError:
    magic_from_file = None

from . import config, gir, log, gdgenerate, utils


HELP_MSG = "Serve the gi-docgen output locally"


class GDHTTPRequestHandler(server.SimpleHTTPRequestHandler):
    SUFFIXES = [
        '.html',
        '/index.html',
        '/',
        '',
    ]

    extensions_map = {
        **server.SimpleHTTPRequestHandler.extensions_map,
        ** {
            ".ttf": "font/ttf",
            ".woff": "font/woff",
            ".woff2": "font/woff2",
        },
    }

    def translate_path(self, path):
        path = path.split('?', 1)[0]
        path = path.split('#', 1)[0]
        trailing_slash = path.rstrip().endswith('/')
        path = urllib.parse.unquote(path)
        path = posixpath.normpath(path)
        words = path.split('/')
        words = filter(None, words)
        path = self.base_path
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # Ignore components that are not a simple file/directory name
                continue
            path = os.path.join(path, word)
        if trailing_slash:
            path += '/'
        return path

    def do_GET(self):
        # cut off a query string
        original_path = self.path.split('?', 1)[0]
        # try to find file
        self.path = self.get_path_that_exists(original_path)
        if not self.path:
            return
        server.SimpleHTTPRequestHandler.do_GET(self)

    def get_path_that_exists(self, original_path):
        # Try to strip trailing slash
        trailing_slash = original_path.endswith('/')
        original_path = original_path.rstrip('/')
        # Try to detect file by applying various suffixes
        tries = []
        for suffix in self.SUFFIXES:
            if not trailing_slash and suffix == '/':
                # if original request does not have trailing slash, skip the '/' suffix
                # so that base class can redirect if needed
                continue
            path = original_path + suffix
            if os.path.exists(self.translate_path(path)):
                return path
            tries.append(path)
        log.warning(f"Unable to find '{original_path}' or variations: {tries}")
        return None

    def guess_type(self, path):
        mimetype = server.SimpleHTTPRequestHandler.guess_type(self, path)
        # If the default guess is too generic, try the python-magic library
        if mimetype == 'application/octet-stream' and magic_from_file:
            mimetype = magic_from_file(path, mime=True)
        return mimetype

    def log_message(self, fmt, *args):
        log.info(fmt % args)


class RootedHTTPServer(server.HTTPServer):
    def __init__(self, base_path, *args, **kwargs):
        server.HTTPServer.__init__(self, *args, **kwargs)
        self.RequestHandlerClass.base_path = base_path


def add_args(parser):
    # Reuse the generate arguments
    gdgenerate.add_args(parser)
    # Add the local server arguments
    parser.add_argument("--bind", "-b", dest="server", default="127.0.0.1",
                        help="The IP used to serve HTTP requests")
    parser.add_argument("--port", "-p", dest="port", type=int, default=8080,
                        help="The port used to serve HTTP requests")


def run(options):
    log.info(f"Loading config file: {options.config}")

    # We need to re-implement gdgenerate.run() here, because the output
    # directory is computed from the configuration and/or the namespace
    # of the introspection data; we could reuse gdgenerate.run(), but it
    # would mean doing the configuration and GIR parsing twice
    conf = config.GIDocConfig(options.config)

    output_dir = options.output_dir or os.getcwd()

    content_dirs = options.content_dirs
    if content_dirs == []:
        content_dirs = [os.getcwd()]

    if options.templates_dir is not None:
        templates_dir = options.templates_dir
    else:
        templates_dir = conf.get_templates_dir()
        if templates_dir is None:
            templates_dir = os.path.join(os.path.dirname(__file__), 'templates')

    theme_name = conf.get_theme_name(default=options.theme_name)
    theme_conf = config.GITemplateConfig(templates_dir, theme_name)

    log.debug(f"Templates directory: {templates_dir}")
    log.info(f"Theme name: {theme_conf.name}")
    log.info(f"Output directory: {output_dir}")

    paths = []
    paths.extend(options.include_paths)
    paths.extend(utils.default_search_paths())
    log.debug(f"Search paths: {paths}")

    log.info("Parsing GIR file")
    parser = gir.GirParser(search_paths=paths)
    parser.parse(options.infile)

    log.checkpoint()
    if options.dry_run:
        return 0

    gdgenerate.gen_reference(conf, options, parser.get_repository(),
                             templates_dir, theme_conf,
                             content_dirs, output_dir)

    if options.no_namespace_dir:
        ns_dir = output_dir
    else:
        namespace = parser.get_repository().namespace
        ns_dir = os.path.join(output_dir, f"{namespace.name}-{namespace.version}")

    RootedHTTPServer.allow_reuse_address = True

    log.checkpoint()
    try:
        httpd = RootedHTTPServer(ns_dir,
                                 (options.server, options.port),
                                 GDHTTPRequestHandler)
    except OSError as e:
        log.error(f"Could not listen on port {options.port}, server {options.server}: {e}")
        return 1
    try:
        log.info(f"Serving site at: http://{options.server}:{options.port} - Press CTRL-C to stop")
        httpd.serve_forever()
    except Exception as e:
        log.warning(f"Error: {e}")
        return 1
    except KeyboardInterrupt:
        log.info("Terminating local server")
        httpd.socket.close()

    return 0



# SPDX-FileCopyrightText: 2021 GNOME Foundation <https://gnome.org>
# SPDX-FileCopyrightText: The Python Markdown Project
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later
# SPDX-License-Identifier: BSD-2-Clause

from markdown.extensions import Extension
from markdown.blockprocessors import BlockProcessor
from markdown.preprocessors import Preprocessor

import xml.etree.ElementTree as etree
import re


SIGNAL_SIGIL_RE = re.compile(r"(^|\W)#([A-Z][A-Za-z0-9]+)::([a-z0-9_-]+)\b")

PROP_SIGIL_RE = re.compile(r"(^|\W)#([A-Z][A-Za-z0-9]+):([a-z0-9_-]+)\b")

TYPE_SIGIL_RE = re.compile(r"(^|\W)#([A-Z][A-Za-z0-9]+)\b")

ARG_SIGIL_RE = re.compile(r"(^|\W)@([A-Za-z0-9_]+)\b")

CONST_SIGIL_RE = re.compile(r"(^|\W)%([A-Z0-9_]+)\b")

FUNCTION_RE = re.compile(r"(^|\s+)([a-z][a-z0-9_]*)\(\)(\s+|$)")


def process_gtkdoc_sigils(line: str) -> str:
    # XXX: The order is important; signals and properties have
    # higher precedence than types
    new_line = line

    # Signal sigil
    new_line = re.sub(SIGNAL_SIGIL_RE, r"\g<1>`\g<2>::\g<3>`", new_line)

    # Property sigil
    new_line = re.sub(PROP_SIGIL_RE, r"\g<1>`\g<2>:\g<3>`", new_line)

    # Type sigil
    new_line = re.sub(TYPE_SIGIL_RE, r"\g<1>`\g<2>`", new_line)

    # Constant sygil
    new_line = re.sub(CONST_SIGIL_RE, r"\g<1>`\g<2>`", new_line)

    # Argument sygil
    new_line = re.sub(ARG_SIGIL_RE, r"\g<1>`\g<2>`", new_line)

    # Function
    new_line = re.sub(FUNCTION_RE, r"\g<1>`\g<2>()`\g<3>", new_line)

    return new_line


class GtkDocPreprocessor(Preprocessor):
    """Remove all gtk-doc sigils from the Markdown text"""
    def run(self, lines):
        new_lines = []
        inside_code_block = False
        for line in lines:
            if line.startswith("```"):
                if not inside_code_block:
                    inside_code_block = True
                else:
                    inside_code_block = False

            new_line = line

            # Never transform code blocks
            if not inside_code_block:
                new_line = process_gtkdoc_sigils(new_line)

            new_lines.append(new_line)
        return new_lines


class GtkDocExtension(Extension):
    """Markdown extension for gtk-doc"""
    def extendMarkdown(self, md):
        """Add extensions"""
        md.preprocessors.register(GtkDocPreprocessor(md), 'gtkdoc', 27)


class AdmonitionProcessor(BlockProcessor):
    """
    The AdmonitionProcessor class from Python-Markdown, modified to
    use a different preamble marker, and with fallback for known
    admonition titles.
    """

    CLASSNAME = 'admonition'
    CLASSNAME_TITLE = 'admonition-title'
    RE = re.compile(r'(?:^|\n)::: ?([\w\-]+(?: +[\w\-]+)*)(?: +"(.*?)")? *(?:\n|$)')
    RE_SPACES = re.compile('  +')
    ADMONITION_TITLE = {
        'note': 'Note',
        'important': 'Important',
        'warning': 'Warning',
        'seealso': 'See also',
        'tip': 'Tip',
        'hint': 'Hint',
    }

    def __init__(self, parser):
        """Initialization."""

        super().__init__(parser)

        self.current_sibling = None
        self.content_indention = 0

    def parse_content(self, parent, block):
        """Get sibling admonition.

        Retrieve the appropriate sibling element. This can get tricky when
        dealing with lists.

        """

        old_block = block
        the_rest = ''

        # We already acquired the block via test
        if self.current_sibling is not None:
            sibling = self.current_sibling
            block, the_rest = self.detab(block, self.content_indent)
            self.current_sibling = None
            self.content_indent = 0
            return sibling, block, the_rest

        sibling = self.lastChild(parent)

        if sibling is None or sibling.tag != 'div' or sibling.get('class', '').find(self.CLASSNAME) == -1:
            sibling = None
        else:
            # If the last child is a list and the content is sufficiently indented
            # to be under it, then the content's sibling is in the list.
            last_child = self.lastChild(sibling)
            indent = 0
            while last_child is not None:
                if (
                    sibling is not None and block.startswith(' ' * self.tab_length * 2)
                    and last_child is not None and last_child.tag in ('ul', 'ol', 'dl')
                ):

                    # The expectation is that we'll find an `<li>` or `<dt>`.
                    # We should get its last child as well.
                    sibling = self.lastChild(last_child)
                    last_child = self.lastChild(sibling) if sibling is not None else None

                    # Context has been lost at this point, so we must adjust the
                    # text's indentation level so it will be evaluated correctly
                    # under the list.
                    block = block[self.tab_length:]
                    indent += self.tab_length
                else:
                    last_child = None

            if not block.startswith(' ' * self.tab_length):
                sibling = None

            if sibling is not None:
                indent += self.tab_length
                block, the_rest = self.detab(old_block, indent)
                self.current_sibling = sibling
                self.content_indent = indent

        return sibling, block, the_rest

    def test(self, parent, block):

        if self.RE.search(block):
            return True
        else:
            return self.parse_content(parent, block)[0] is not None

    def run(self, parent, blocks):
        block = blocks.pop(0)
        m = self.RE.search(block)

        if m:
            if m.start() > 0:
                self.parser.parseBlocks(parent, [block[:m.start()]])
            block = block[m.end():]  # removes the first line
            block, theRest = self.detab(block)
        else:
            sibling, block, theRest = self.parse_content(parent, block)

        if m:
            klass, title = self.get_class_and_title(m)
            div = etree.SubElement(parent, 'div')
            div.set('class', '{} {}'.format(self.CLASSNAME, klass))
            if title:
                p = etree.SubElement(div, 'p')
                p.text = title
                p.set('class', self.CLASSNAME_TITLE)
        else:
            # Sibling is a list item, but we need to wrap it's content should be wrapped in <p>
            if sibling.tag in ('li', 'dd') and sibling.text:
                text = sibling.text
                sibling.text = ''
                p = etree.SubElement(sibling, 'p')
                p.text = text

            div = sibling

        self.parser.parseChunk(div, block)

        if theRest:
            # This block contained unindented line(s) after the first indented
            # line. Insert these lines as the first block of the master blocks
            # list for future processing.
            blocks.insert(0, theRest)

    def get_class_and_title(self, match):
        klass, title = match.group(1).lower(), match.group(2)
        klass = self.RE_SPACES.sub(' ', klass)
        if title is None:
            if klass in self.ADMONITION_TITLE:
                # Use one of the known titles, for consistency
                title = self.ADMONITION_TITLE[klass]
            else:
                # no title was provided, use the capitalized class name as title
                # e.g.: `::: foo` will render
                # `<p class="admonition-title">Foo</p>`
                title = klass.split(' ', 1)[0].capitalize()
        elif title == '':
            # an explicit blank title should not be rendered
            # e.g.: `!!! warning ""` will *not* render `p` with a title
            title = None
        return klass, title


class AdmonitionExtension(Extension):
    """Notice extension"""
    def extendMarkdown(self, md):
        """Add extensions"""
        md.registerExtension(self)
        md.parser.blockprocessors.register(AdmonitionProcessor(md.parser), 'admonition', 105)
